"""
GIT repository push management
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import re
import git


class GitPush:
    """Handles pushing to a git repo"""
    __ssh_prefix = 'git@'

    def __init__(self, repo, credentials=None):
        """Initialize this repository

        :param repo: git repository
        :param credentials: git credential (optional, defaults to None)
        """
        self._repo = repo
        self._credentials = credentials
        self._credentials_set = self._credentials is None

    def push_to_origin(self, item):
        """Push the item to origin

        :param item: item to push
        """
        self._push_possibly_adding_credentials_and_trying_again_on_authentication_error(item)

    def _push_possibly_adding_credentials_and_trying_again_on_authentication_error(self, item):
        while True:
            try:
                self._repo.remotes.origin.push(item)
                break
            except git.GitCommandError as ex:
                if '403' not in ex.stderr:
                    raise
                self._set_credentials_once_for_this_instance()

    def _set_credentials_once_for_this_instance(self):
        if self._credentials_set:
            raise ConnectionRefusedError(
                f'Unable to connect with {"specified" if self._credentials else "no"} credentials'
                f' to {self._repo}')
        self._enable_credentials()

    def _enable_credentials(self):
        if not self._ssh_urls_configured_for_repo_origin():
            self._configure_ssh_url_for_repo_origin()
        self._credentials.enable()

    def _ssh_urls_configured_for_repo_origin(self):
        return [url for url in
                self._repo.remotes.origin.urls if url.startswith(self.__ssh_prefix)]

    def _configure_ssh_url_for_repo_origin(self):
        http_url, server, project = self._get_http_url_server_and_project()
        ssh_url = self._make_ssh_url(server, project)
        self._repo.remotes.origin.set_url(ssh_url, http_url)

    def _get_http_url_server_and_project(self):
        pattern = re.compile(r'http.?://([^/]+)/(.*)')
        matches = [match for match in
                   [(url, pattern.match(url)) for url in self._repo.remotes.origin.urls]
                   if match[1]]
        if not matches:
            raise IOError('No HTTP* URLs found in repo.origin')
        first_match = matches[0]
        return (first_match[0], ) + tuple(first_match[1].groups())

    @staticmethod
    def _make_ssh_url(server, project):
        return f'{GitPush.__ssh_prefix}{server}:{project}'



"""
GIT tag-based version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import re
from pykrete.args import CiIo
from pykrete.repo import Git
from .revision_type import RevisionType
from .version import Version
from .ci_revision import CiRevision


class TagVersion(Version):
    """Handle version in CI environment
    A version is constructed from the last version tag (tag name format '<major>.<minor>.<build>'
     indicating a release version, with a message format of 'ci_base_build <build>' indicating the
     CI build number for the first release with this major-minor combo). In the case of a release,
     any following tag named 'bump_major' will increment the major version (at most once), otherwise
     any following tag named 'bump_minor' will increment the minor version (at most once).
    Build is calculated as the current CI project-build# variable, minus the build specified in the
     last version tag's message if this isn't a release, or 0 if it is a release.
    Revision is read from the CI environment variables CI_COMMIT_REF_NAME, CI_MERGE_REQUEST_TITLE
     and CI_JOB_NAME:
        Release (4) - master branch on a job who'se name doesn't contain '_rc_'
        RC (3) - non-WIP merge request, master branch on a job who'se name contains '_rc_'
        Beta (2) - merge request with 'WIP' in the title
        Alpha (1) - none of the above
    """
    def __init__(self, repo=Git(), ci_io=None):
        """Initializes this instance from CI environment

        :param repo: project's repository (optional, defaults to git for the current folder)
        :param ci_io: CI environment's IO manager (optional, defaults to pykrete.args.CiIo)
        """
        self.__ci_io = ci_io if ci_io else CiIo()
        self.__repo = repo
        self.__tags = self.__read_tags()
        self.__revision = CiRevision(self.__ci_io)
        self.__base_build = 0
        (major, minor, build) = self.__read_version_from_tags()
        super().__init__(major=major,
                         minor=minor,
                         revision=self.__revision.revision_type.value,
                         build=build,
                         revision_type=self.__revision.revision_type)

    def __read_tags(self):
        return self.__repo.get_tags_from(pattern=r'(\d+\.){2}\d+', is_must=False)

    def __read_version_from_tags(self):
        last_version_tag = self.__tags[0] if self.__tags else None
        version_parts, self.__base_build =\
            self.__get_version_parts_and_base_build_from(last_version_tag)
        version_parts = self.__advance_version(version_parts)
        return tuple(version_parts)

    @staticmethod
    def __get_version_parts_and_base_build_from(tag):
        if not tag:
            return [0, 0, 0], 0
        version_parts = [int(part) for part in tag[0].split('.')]
        message_build = re.findall(r'ci_base_build (\d+)', tag[1]) if tag[1] else None
        base_build = int(message_build[0]) if message_build else 0
        return version_parts, base_build

    def __advance_version(self, version_parts):
        ci_build = int(self.__ci_io.read_env('build version'))
        if self.__revision.revision_type != RevisionType.Release:
            return self.__advance_build_relative_to_base_build(version_parts, ci_build)
        return self.__advance_major_minor_from_last_change_details(version_parts, ci_build)

    def __advance_build_relative_to_base_build(self, version_parts, ci_build):
        version_parts[2] += ci_build - self.__base_build
        return version_parts

    def __advance_major_minor_from_last_change_details(self, version_parts, ci_build):
        change_details = self.__repo.get_head_change().details
        if '#major' in change_details:
            return [version_parts[0] + 1, 0, 0]
        if '#minor' in change_details:
            return [version_parts[0], version_parts[1] + 1, 0]
        return self.__advance_build_relative_to_base_build(version_parts, ci_build)



"""
GIT repository management
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging

import git
from pykrete.repo.change import Change
from pykrete.repo.repo import Repo
from .git_internals.git_push import GitPush
from .git_internals.git_tags import GitTagsReader


class Git(Repo):
    """Handles a GIT repository"""

    def __init__(self, path='.', credentials=None):
        """Initialize this repository

        :param path: path to repository (optional, defaults to the CWD)
        :param credentials: git credential (optional, defaults to None)
        """
        self._logger = logging.getLogger(__name__)
        self._repo = git.Repo.init(path)
        self._pusher = GitPush(self._repo, credentials)
        self._tag_reader = GitTagsReader(self._repo)

    def add_tag(self, name, message):
        """Adds a tag to the repo

        :param name: tag name
        :param message: tag message
        :return: the new tag's info tuple (name, message)
        """
        new_tag = self._repo.create_tag(name, message=message)
        self._pusher.push_to_origin(new_tag)
        return self._tag_reader.name_message_tuple_from_tag_ref(new_tag)

    def remove_tag(self, name, is_must=True):
        """Adds a tag to the repo

        :param name: tag name
        :param is_must: True to raise an error tag doesn't exist, False to just skip in that case.
        :exception: KeyError - tag not found
        """
        old_tags = self._tag_reader.get_last_tag_named(name, is_must)
        old_tag = old_tags[0]
        self._repo.delete_tag(old_tags)
        self._pusher.push_to_origin(old_tag)

    def get_all_tags(self):
        """Gets all tags from the current branch in the repo

        :return: A list of tag info tuples (name, message)
        """
        return GitTagsReader(self._repo).tags

    def get_tags_from(self, pattern, is_must=False):
        """Gets all tags from the current branch in the repo, from the tag whose name matches the
         supplied pattern.

        :param pattern: name pattern to match
        :param is_must: True to raise an error if no match is found, False to return all tags in
        that case (with the first tuple as None).
        :return: A list of tag info tuples (name, message)
        :exception: KeyError - tag not found
        """
        return self._tag_reader.get_tags_from(pattern, is_must)

    def get_head_change(self, branch=None):
        """Gets the last change in the branch

        :param branch: branch name (optional, defaults to the current branch)
        :return: (pykrete.repo.Change) last change in the branch
        """
        commit = self._repo.commit(branch)
        return self.__append_commits_to_change([commit])

    @staticmethod
    def __append_commits_to_change(commits, change=Change()):
        for commit in commits:
            if commit:
                if commit.summary:
                    change.append_log(commit.summary)
                if commit.message:
                    change.append_details(commit.message)
        return change



"""
Pykrete repo testbed
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

from pykrete.repo import Repo


class RepoTestbed(Repo):
    """Testbed class for tests which need a simulated repo"""

    def __init__(self, tags, change):
        """Initialize the testbed

        :param tags: simulated tags
        :param change: simulated change
        """
        self.tags = tags
        self.change = change

    def get_all_tags(self):
        """simulation"""
        return self.tags

    def get_tags_from(self, pattern, is_must=False):
        """simulation"""
        if is_must:
            raise AssertionError('tags are fetched with is_must=True')
        if not pattern:
            raise AssertionError('no pattern was provider for fetching tags')
        return self.tags

    def get_head_change(self, branch=None):
        """simulation"""
        return self.change

    def add_tag(self, name, message):
        """simulation"""
        return 'some_tag', 'some_message'

    def remove_tag(self, name, is_must=True):
        """simulation"""



"""
Environment argument handling
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import os


def environ(var_name, role=None):
    """Gets an environment variable's value

    :param var_name: The variable's name
    :param role: The variable's role (optional)
    :return: The variable's value
    :exception SystemError: The variable is not defined in the environment and role is specified
    """
    try:
        return os.environ[var_name]
    except KeyError:
        if role:
            raise SystemError(f'Build requires that the "{var_name}"'
                              f' environment variable is set to {role}')
        return None



"""
Python package
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from pkg_resources import require
from pykrete.versioning import Formatting, TagVersion, VersionPyVersion


class PythonPackage:
    """Python package"""

    @property
    def project(self):
        """
        :return: (string) name of project package under 'src'.
        """
        return self._project

    @property
    def version(self):
        """
        :return: (string) The project's version
        """
        return self._version

    def __init__(self, project, package_fallback=False):
        """Initializes this instance to analyze the specified project

        :param project: name of project package under 'src'
        """
        self._project = project
        self._package_fallback = package_fallback
        self._version = self._get_version()

    @property
    def long_description(self):
        """Gets the contents of the README.md file

        :return: Long description
        """
        with open('README.md', 'r') as readme:
            return readme.read()

    def _get_version(self):
        """Gets the package version

        :return: The version
        :exception IndexError: Version not found
        :exception IOError: Version read failed
        """
        try:
            version = TagVersion()
        except SystemError:
            try:
                version = VersionPyVersion(self._project)
            except IOError:
                if not self._package_fallback:
                    raise
                version_from_current_package = require(self._project)[0].version
                return version_from_current_package
        return Formatting(version).for_python

    def __str__(self):
        return f'{self._project} v{self._version}'



"""
Pykrete Args tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import logging
import unittest
import argparse
from pykrete.args import build_version, exiting_file, yes_or_no


class PykreteArgsTestCase(unittest.TestCase):
    """Unit tests for pykrete's args module
    """

    _logger = logging.getLogger(__name__)

    def test_build_version_good(self):
        """Verifies correct parsing of a legal version
        """
        source = '1.2.3.4'
        version = build_version(source)
        self.assertEqual(version.string, source, 'read version differs from source')
        self._logger.debug(':'.join(version.groups()[1:4]))

    def test_build_version_short(self):
        """Verifies rejection of a short version
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(build_version('1.2.3'))

    def test_build_version_bad(self):
        """Verifies rejection of a non-version
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(build_version('Blue. No, yellOOOOOOOW!!'))

    def test_existing_file_good(self):
        """Verifies acceptance of an existing file
        """
        source = 'README.md'
        self.assertEqual(exiting_file(source), source, 'read file name differs from source')

    def test_existing_file_bad(self):
        """Verifies rejection of a non-existing file
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(exiting_file('no_such_file.txt'))

    def test_yes_no_good(self):
        """Verifies correct parsing of yes/no answers
        """
        cases = {True: ['yes', 'Yes', 'YES'], False: ['no', 'No', 'NO']}
        for key, sources in cases.items():
            for source in sources:
                self.assertEqual(key, yes_or_no(source), f'{source} not {key}')

    def test_yes_no_bad(self):
        """Verifies rejection of a non-yes/no answer
        """
        with self.assertRaises(argparse.ArgumentTypeError):
            self._logger.debug(yes_or_no('Huh? I don\'t know that! EEEEEEAAAARRGH!!!'))


if __name__ == '__main__':
    unittest.main()



"""
CI environment-based version information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from pykrete.args import CiIo
from .version import Version
from .ci_revision import CiRevision


class CiVersion(Version):
    """Handle version in CI environment
    Version parts are read from CI variables CI_VERSION_MAJOR, CI_VERSION_MINOR and CI_PIPELINE_IID
    Revision is read from CI_COMMIT_REF_NAME, CI_MERGE_REQUEST_TITLE and CI_JOB_NAME:
        Release - in master branch and job's name doesn't contain '_rc_'
        RC - in non-WIP merge request, or in master branch with a job who'se name contains '_rc_'
        Beta - in merge request with 'WIP' in the title
        Alpha - none of the above
    """
    def __init__(self, ci_io=None):
        """Initializes this instance from CI environment

        :param ci_io: CI environment's IO manager (optional, defaults to pykrete.args.CiIo)
        """
        self.__ci_io = ci_io if ci_io else CiIo()
        revision_type = CiRevision(self.__ci_io).revision_type
        super().__init__(major=self.__ci_io.read_env('major version'),
                         minor=self.__ci_io.read_env('minor version'),
                         revision=revision_type.value,
                         build=self.__ci_io.read_env('build version'),
                         revision_type=revision_type)



"""
Python build manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import subprocess
import shutil
import datetime
import logging
from pykrete.packages import install_python_packages
from pykrete.args import environ
from pykrete.versioning import Formatting, make_python_root_version_file_path
from .build_manager import BuildManager


class PythonBuildManager(BuildManager):
    """Manages building a python package"""

    _logger = logging.getLogger(__name__)

    def _clean_distribution(self):
        """Removes previous distribution files"""
        self._logger.debug('Deleting previous distribution')
        shutil.rmtree('dist', ignore_errors=True)

    def _upload(self):
        """Uploads the package if self.upload is True"""
        install_python_packages('twine')
        self._make_pypirc()
        self._run_setup(
            'Uploading package...',
            'twine', '--project', self._project)

    def _build(self):
        """Performs the actual build using setup.py"""
        install_python_packages('setuptools', 'wheel')
        self._run_setup(
            'Building package...',
            'check', 'sdist', 'bdist_wheel')

    def _set_package_version(self):
        """Sets the package version in the project's version file"""
        version = Formatting(self._version).for_python
        self._logger.debug('Setting version to: %s', version)
        with open(make_python_root_version_file_path(self._project), 'w') as version_file:
            self._write_lines(version_file, self._make_version_file_lines(version))

    def _make_version_file_lines(self, version):
        """Makes the lines of version.py

        :return: The version file's lines
        """
        return [
            '"""',
            f'{self._project.title()} module version',
            'AUTO-GENERATED BY BUILD.PY',
            f'(C) {datetime.date.today().year}',
            '"""',
            '',
            f'__version__ = \'{version}\'',
            ''
        ]

    def _run_setup(self, message, *args):
        """Runs setup.py

        :param message: pre-run debug message
        :param args: setup.py argument
        """
        self._logger.debug(message)
        subprocess.check_call(['python', 'setup.py'] + list(args))

    def _make_pypirc(self):
        """Generates the pypirc file required for upload, from GitLab environment variables:
        username: PYPI_USER
        password: PYPI_PASSWORD
        """
        self._logger.debug('Prepping pypirc')
        path = environ('pypirc')
        with open(path if path else 'pypirc', 'w') as pypirc:
            self._write_lines(pypirc, self._make_pypirc_lines())

    @staticmethod
    def _write_lines(file, lines):
        file.writelines('\n'.join(lines) + '\n')

    @staticmethod
    def _make_pypirc_lines():
        return [
            '[distutils]',
            'index-servers =',
            '    pypi',
            '',
            '[pypi]',
            '    username: ' + environ('PYPI_USER', 'the Pypi repository user name'),
            '    password: ' + environ('PYPI_PASSWORD', 'the Pypi repository password/API-key'),
        ]



"""
distutils project-command base
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from abc import abstractmethod
from distutils.cmd import Command


class ProjectCommand(Command):
    """A custom command base for project commands
    """

    _project_option = ('project=', None, 'Name of project package under \'src\'')
    user_options = [_project_option]

    def __init__(self, dist):
        """
        :param dist: distribution
        """
        self.project = None
        super().__init__(dist)

    def initialize_options(self):
        """Set default values for options"""
        self.project = None

    def finalize_options(self):
        """Post-process received options"""
        assert self.project, 'project name not specified'

    @abstractmethod
    def run(self):
        """Performs the command"""



"""
CI information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .environ import environ


class CiIo:
    """Handle CI environment IO
    """
    def __init__(self, ci_reader=environ, ci_spec=None):
        """Initializes this instance from CI environment

        :param ci_reader: (ci_variable, role)=>value function [see pykrete.args.environ behavior]
        :param ci_spec: dictionary of part to CI variable
         (parts are: 'major version', 'minor version', 'build version', 'branch name',
          'merge request title', 'job name', 'deploy key')
        """
        self.__ci_reader = ci_reader
        self.__ci_spec = ci_spec if ci_spec else self.__get_default_ci_spec()

    def read_env(self, part, is_no_role=False):
        """Read a spec-part from the CI environment

        :param part: part name
        :param is_no_role: should there be an exception if not found
        :return: Read value, or None if not found
        :exception: The part's variable is not defined in the environment and it should be
        """
        return self.__ci_reader(self.__ci_spec[part],
                                None if is_no_role else f'CI\'s \'{part}\' value')

    @staticmethod
    def __get_default_ci_spec():
        return {'major version': 'CI_VERSION_MAJOR',
                'minor version': 'CI_VERSION_MINOR',
                'build version': 'CI_PIPELINE_IID',
                'branch name': 'CI_COMMIT_REF_NAME',
                'merge request title': 'CI_MERGE_REQUEST_TITLE',
                'job name': 'CI_JOB_NAME',
                'deploy key': 'CI_DEPLOY_KEY'}



"""
Pykrete setup
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

from .project_test_command import ProjectTestCommand
from .self_test_command import SelfTestCommand

__all__ = ['ProjectTestCommand', 'SelfTestCommand']



"""
Pykrete arguments
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

from .args import build_version, exiting_file, yes_or_no
from .environ import environ
from .ci_io import CiIo

__all__ = ['build_version', 'exiting_file', 'yes_or_no', 'environ', 'CiIo']



"""
distutils pylint-command base
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from os import path
from distutils.log import ERROR
from distutils.errors import DistutilsModuleError
from .project_command import ProjectCommand
from ..calls import CheckedCall
from ..packages import PythonPackage


class TwineCommand(ProjectCommand):
    """A custom command to upload the package to pypi"""

    description = 'Upload the project to pypi'
    user_options = [ProjectCommand._project_option,
                    ('index=', None, 'Index server (optional, defaults to pipy)'),
                    ('config=', None, 'pypirc location (optional, defaults to local file)')]

    def __init__(self, dist):
        """
        :param dist: distribution
        """
        self.__package = None
        self.index = None
        self.config = None
        super().__init__(dist)

    def initialize_options(self):
        """Set default values for options"""
        self.index = 'pypi'
        self.config = 'pypirc'

    def finalize_options(self):
        """Post-process received options"""
        self.__package = path.join(
            'dist',
            f'{self.project}-{PythonPackage(self.project).version}.tar.gz')

    def run(self):
        """Register and upload the package

        :exception OSError: upload failed
        """
        self._twine('upload')

    def _twine(self, command):
        """Run a twine command

        :param command: A twine command
        :exception OSError: command failed
        """
        cmd = CheckedCall(['twine', command, '--config-file', self.config,
                           '-r', self.index, self.__package])
        if not cmd.success:
            self.announce(cmd.get_output(), ERROR)
            raise DistutilsModuleError(f'Twine {command} failed.')



"""
Version information managers
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .ci_version import CiVersion
from .revision_type import RevisionType
from .version import Version
from .version_py_version import VersionPyVersion, make_python_root_version_file_path
from .tag_version import TagVersion
from .formatting import Formatting

__all__ = ['CiVersion', 'RevisionType', 'Version', 'VersionPyVersion', 'TagVersion', 'Formatting',
           'make_python_root_version_file_path']



"""
Pykrete repo tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
import logging
import uuid
from pykrete.repo import Git
from .helpers import get_full_path


class PythonRepoTestCase(unittest.TestCase):
    """Unit tests for pykrete's repo module
    """

    _logger = logging.getLogger(__name__)

    def test_get_all_tags(self):
        """Verifies getting all tags from repo
        """
        self._assert_get_tags(lambda git: git.get_all_tags(),
                              ('1.0.0', None), ('1.0.1', 'bug fix'))

    def test_get_tags_from_existing(self):
        """Verifies getting all tags from repo starting from a specific tag
        """
        self._assert_get_tags(lambda git: git.get_tags_from(r'.*?0\.1', True),
                              ('1.0.1', 'bug fix'))

    def test_get_tags_from_must_but_non_existing(self):
        """Verifies getting all tags from repo
        """
        with self.assertRaises(KeyError):
            self._assert_get_tags(lambda git: git.get_tags_from(r'no such tag', True))

    def test_get_tags_from_non_must_non_existing(self):
        """Verifies getting all tags from repo
        """
        self._assert_get_tags(lambda git: git.get_tags_from(r'no such tag', False),
                              None, ('1.0.0', None), ('1.0.1', 'bug fix'))

    def test_get_head_change(self):
        """Verifies a non-empty head-change returned from repo"""
        target = self._make_target('.')
        change = target.get_head_change()
        self.assertIsNotNone(change, 'empty change returned')
        self._logger.debug(change)
        self.assertIsNotNone(change.log, 'empty log returned')
        self.assertIsNotNone(change.details, 'empty log returned')

    def test_add_remove_tag_fails_with_no_credentials(self):
        """Verifies adding and removing tag in the repo"""
        target = self._make_target(get_full_path('testing-grounds', True))
        expected = (str(uuid.uuid4()), 'test add remove tag')
        with self.assertRaises(ConnectionRefusedError):
            new_tag = target.add_tag(expected[0], expected[1])
            self.assertEqual(expected, new_tag, 'new tag not as expected')
            self.assertTrue(expected in target.get_all_tags(), 'new tag not found')
            target.remove_tag(expected[0])
            self.assertFalse(expected in target.get_all_tags(), 'removed tag still found')

    def test_remove_non_existing_tag(self):
        """Verifies adding and removing tag in the repo"""
        target = self._make_target(get_full_path('testing-grounds', True))
        with self.assertRaises(KeyError):
            target.remove_tag(str(uuid.uuid4()))

    def _assert_get_tags(self, getter, *expected):
        """Verifies tags from repo

        :param getter: tag getter (lambda on Git object)
        :param expected: expected returned tags
        :return:
        """
        target = self._make_target()
        tags = getter(target)
        self._logger.debug(tags)
        self.assertTrue(tags, 'Got empty tags')
        self.assertTrue(len(tags) >= len(expected), 'too few tags returned')
        for test in zip(expected, tags):
            self.assertEqual(test[0], test[1], 'wrong tag returned')

    @staticmethod
    def _make_target(path=None):
        test_repo_path = path if path else get_full_path('git-tag-example', True)
        return Git(test_repo_path)


if __name__ == '__main__':
    unittest.main()



"""
State machine's entry point
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .states import BetweenArguments
from .parsing_state import ParsingState
from .parse_context import ParseContext


def get_starting_state() -> ParsingState:
    """
    :return: state at start position
    """
    return BetweenArguments(ParseContext())



"""
CI environment-based revision information manager
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .revision_type import RevisionType


class CiRevision:
    """Handle revision in CI environment
    Revision is read from CI_COMMIT_REF_NAME, CI_MERGE_REQUEST_TITLE and CI_JOB_NAME:
        Release (4) - master branch on a job who'se name doesn't contain '_rc_'
        RC (3) - non-WIP merge request, master branch on a job who'se name contains '_rc_'
        Beta (2) - merge request with 'WIP' in the title
        Alpha (1) - none of the above
    """
    @property
    def revision_type(self):
        """
        :return: Revision type of the version
        """
        return self._revision_type

    def __init__(self, ci_io):
        """Initializes this instance from CI environment

        :param ci_io: CI IO for 'branch name', 'merge request title'
        """
        self.__ci_io = ci_io
        self._revision_type = self.__get_revision_type()

    def __get_revision_type(self):
        if self.__is_master():
            return self.__revision_from_master()

        revision = self.__revision_from_merge_request()
        return revision if revision else RevisionType.Alpha

    def __is_master(self):
        branch_name = self.__ci_io.read_env('branch name')
        return branch_name == 'master'

    def __revision_from_merge_request(self):
        merge_request_title = self.__ci_io.read_env('merge request title', True)
        if merge_request_title:
            return RevisionType.Beta if 'WIP' in merge_request_title else RevisionType.RC
        return None

    def __revision_from_master(self):
        job_name = self.__ci_io.read_env('job name')
        return RevisionType.RC if '_rc_' in job_name else RevisionType.Release



"""
Pykrete module build script
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
# pylint: disable=wrong-import-order
import src_appender
from logging.config import fileConfig
from setuptools import setup, find_packages
from pykrete.packages import PythonPackage
from pykrete.distrib import TwineCommand
from pykrete.distrib.pylint import SelfTestCommand, ProjectTestCommand


fileConfig('loggingrc')
src_appender.print_path()
PACKAGE = PythonPackage('pykrete')
print(f'Handling pip package: {PACKAGE}')
setup(
    name='pykrete',
    version=PACKAGE.version,
    license='MIT',
    author='Shai A. Bennathan',
    author_email='shai.bennathan@intel.com',
    description='Build script foundation',
    long_description=PACKAGE.long_description,
    long_description_content_type='text/markdown',
    url='http://ait-tool-center.iil.intel.com/',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    test_suite='nose.collector',
    tests_require=['nose'],
    cmdclass={
        'pylint_self': SelfTestCommand,
        'pylint': ProjectTestCommand,
        'twine': TwineCommand
    }
    # install_requires=[],
)



"""
Pykrete versioning.ci_version tests
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

import unittest
import logging
from pykrete.args import CiIo
from pykrete.versioning import CiVersion, RevisionType
from .versioning import PykreteVersioningTestCase


class PykreteVersioningCiVersionTestCase(PykreteVersioningTestCase):
    """Unit tests for pykrete's versioning module's CiVersion class
    """

    _logger = logging.getLogger(__name__)

    def test_ci_version_server(self):
        """Verifies handling of CI-environment version
        NOTE: THIS TEST WILL FAIL LOCALLY, SAFE TO IGNORE
        """
        self._logger.info('NOTE: THIS TEST WILL FAIL LOCALLY, SAFE TO IGNORE')
        self._assert_version_pattern(CiVersion())

    def test_ci_version_local_release(self):
        """Verifies handling of simulated CI-environment's release version"""
        self._assert_ci_version_local(
            spec={'major version': '1',
                  'minor version': '2',
                  'build version': '3',
                  'branch name': 'master',
                  'merge request title': 'None',
                  'job name': 'some_job'},
            expected_revision_type=RevisionType.Release,
            expected_revision=4)

    def test_ci_version_local_rc_master_rc_job(self):
        """Verifies handling of simulated CI-environment's release-candidate version"""
        self._assert_ci_version_local(
            spec={'major version': '11',
                  'minor version': '22',
                  'build version': '33',
                  'branch name': 'master',
                  'merge request title': 'None',
                  'job name': 'some_rc_job'},
            expected_revision_type=RevisionType.RC, expected_revision=3)

    def test_ci_version_local_rc_merge_request(self):
        """Verifies handling of simulated CI-environment's release-candidate version"""
        self._assert_ci_version_local(
            spec={'major version': '12',
                  'minor version': '23',
                  'build version': '34',
                  'branch name': 'custom',
                  'merge request title': 'title'},
            expected_revision_type=RevisionType.RC, expected_revision=3)

    def test_ci_version_local_beta(self):
        """Verifies handling of simulated CI-environment's beta version"""
        self._assert_ci_version_local(
            spec={'major version': '111',
                  'minor version': '222',
                  'build version': '333',
                  'branch name': 'custom',
                  'merge request title': 'WIP: title'},
            expected_revision_type=RevisionType.Beta, expected_revision=2)

    def test_ci_version_local_alpha(self):
        """Verifies handling of simulated CI-environment's alpha version"""
        self._assert_ci_version_local(
            spec={'major version': '123',
                  'minor version': '456',
                  'build version': '789',
                  'branch name': 'custom',
                  'merge request title': 'None'},
            expected_revision_type=RevisionType.Alpha,
            expected_revision=1)

    def _assert_ci_version_local(self, spec, expected_revision_type, expected_revision):
        target = CiVersion(CiIo(self._echo, spec))
        self._assert_version_pattern(target)
        self._assert_spec(spec, target)
        self.assertEqual(expected_revision_type, target.revision_type, "Wrong revision type")
        self.assertEqual(expected_revision, target.revision, "Wrong revision")

    def _assert_spec(self, spec, target):
        self.assertEqual(spec['major version'], target.major, "Wrong major version")
        self.assertEqual(spec['minor version'], target.minor, "Wrong minor version")
        self.assertEqual(spec['build version'], target.build, "Wrong build version")


if __name__ == '__main__':
    unittest.main()



"""
Pykrete versioning tests base
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import re
import os
import unittest
import logging


def _echo(what, who):
    print(f'CiVersion is reading \'{who}\' and will get \'{what}\'')
    return None if what == 'None' else what


def _get_full_path(project):
    return os.path.join('pykrete', 'tests', 'src', project)


class PykreteVersioningTestCase(unittest.TestCase):
    """Unit tests base for pykrete's versions module
    """

    _logger = logging.getLogger(__name__)
    __version_patten = re.compile(r'(\d+\.){3}\d+ (Alpha|Beta|RC|Release)')

    def _assert_version_pattern(self, target):
        self._logger.debug('tested version: %s', target)
        self.assertTrue(self.__version_patten.match(str(target)), 'Wrong version format')

    def _echo(self, what, who):
        self._logger.debug('CiVersion is reading \'%s\' and will get \'%s\'', who, what)
        return None if what == 'None' else what
